#!/usr/bin/perl
#        .
#     : 
#      -      ;
#      -  ;
#      -   .
# v0.2.4 Copyright (c) 2003-2004 by Maxim Chirkov <mc@tyumen.ru>

use BerkeleyDB;
use Socket;

require "/etc/spam_check/spam_check.conf";

if ($cfg_log_watch_mode == 1){
    # "tail"    .
    use File::Tail;
}

my $sendmail_format;
if ($cfg_log_format == 1) {
    $sendmail_format='ERROR:';
} else {
    $sendmail_format='';
}

my %chk_cache;
my $db_obj;

#  .
# Fix for Broken pipe.
sub pipe_sig{ 
    return 0; 
};
local ($SIG{PIPE}) = \&pipe_sig;

#     .
sub hup_sig{ 
    $db_obj->db_sync();
    return 0; 
};
local ($SIG{HUP}) = \&hup_sig;

#     .
sub kill_sig{ 
    unlink($cfg_lock_file);
    $db_obj->db_sync();
    untie %chk_cache;
    print STDERR "SIGKILL ! Flushing buffers and aborting...";
    exit(0);
};
local ($SIG{INT}) = \&kill_sig;
local ($SIG{TERM}) = \&kill_sig;


#     .
my $now_time = time();
#my $today_time = int($now_time/(24*60*60));

#    IP    .
my %chk_ip = ();
my %rejected_ip = ();
my %active_session=();
my $host_cnt=0; #       .

#------------------------------------------------------------

    #   .
    if (-f "$cfg_lock_file"){
	if ((stat($cfg_lock_file))[9] + $cfg_lock_timetolive < $now_time){
	    sys_log(0, "WARN:Lock removed.");
	    unlink($cfg_lock_file);
	} else {
	    sys_log(0, "WARN:Lock file exists.");
	    die "WARN:Lock file exists.";
	}
    }
    #   .
    open(LOCK, ">$cfg_lock_file") or sys_log (-1, "ERROR:Can't create lock file $cfg_lock_file");
    flock(LOCK, 2);
    print LOCK "$$\n";
    close(LOCK);

    #          .
    my @local_ip_bin_list = ();
    optimize_ip_array(\@cfg_local_ip_list, \@local_ip_bin_list);

    #          .
    if ((-f "$cfg_active_mail_file") 
         && (open (A_SES, "<$cfg_active_mail_file"))){
	flock(A_SES, 1);
	while(<A_SES>){
	    chomp;
	    my ($cur_ip, $cur_hostname, $cur_time) = split(/\t/);
	    if ($cur_time + $cfg_lock_timetolive > $now_time){
	        $active_session{$cur_ip} = $cur_hostname;
	    } else {
		sys_log(2, "WARN: Dissconnect record not found for $cur_ip:$cur_hostname");
	    }
	}
	close(A_SES);
    } else {
    	sys_log(2, "INFO: $cfg_active_mail_file not found, new iteration");
    }

    if (! -f $cfg_maillog_file){
	sys_log(-1, "ERROR:Maillog not found ($cfg_maillog_file)");
    }

    #     .
    my ($last_log_position, $last_log_size);
    my ($new_log_position, $new_log_size);
    
    if ($cfg_log_watch_mode == 0){ # "cron"    .
	
	if (-f "$cfg_seek_file"){
	    load_last_position(\$last_log_position, \$last_log_size, $cfg_seek_file);
	} else {
	    #  .
	    sys_log(2, "INFO: First start. File with log checkpoint not found. Setting checkpoint to end of file.");
	    my $new_log_size = (stat($cfg_maillog_file))[7];
    	    save_cur_position($new_log_size, $new_log_size, $cfg_seek_file);
    	    unlink($cfg_lock_file);
	    exit(0);
	}
	$new_log_size = (stat($cfg_maillog_file))[7];

	if ($new_log_size < $last_log_size){
	    #   .
            $last_log_position = 0;
	    sys_log(2, "INFO: Maillog after rotation, seek to begin.");
        } elsif ($new_log_size == $last_log_size){
		#   .
    	    sys_log(2, "INFO: Maillog have't changed from last check. Exiting.");
    	    unlink($cfg_lock_file);
	    exit(0);
	}
    }

    #       .
    $db_obj = tie (%chk_cache, 'BerkeleyDB::Btree',
	-Filename   => $cfg_checker_cache_file,
	-Flags      => DB_CREATE,
	-Cachesize   => DB_CACHE_SIZE)  or sys_log(-1, "ERROR:Cannot create/open $cfg_checker_cache_file: $!");

# --------------------------
    #    .
    if ($cfg_log_watch_mode == 0){ # "cron"    .
	open(MAILLOG, "<$cfg_maillog_file") or sys_log(-1, "ERROR:Can't open maillog $cfg_maillog_file");
	if ($last_log_position > 0){
	    if (seek(MAILLOG, $last_log_position, 0) != 1){
		sys_log(0, "ERROR:seek() return error: $!");
	    }
	}
    }elsif ($cfg_log_watch_mode == 1) { # "tail"    .
	my $tail_ref = tie *MAILLOG, "File::Tail", ( name=>$cfg_maillog_file, 
	    maxinterval=> $cfg_blocklist_rebuild_interval,
	    interval=> $cfg_blocklist_rebuild_interval) or sys_log(-1, "ERROR:Can't open maillog $cfg_maillog_file");
    }
    while (my $cur_line = <MAILLOG>){
	if ($cur_line =~ /^.*$cfg_maillog_connect_mask$/){
	    sys_log(3, "DEBUG:CONNECT Line: $cur_line");

	    #  .
	    my $cur_hostname=$1;
	    my $cur_ip=$2;
	    if (check_ip_in_list(\@local_ip_bin_list, $cur_ip) == 1){
		#   IP.
		sys_log(2, "INFO: $cur_ip/$cur_hostname defined as local network (cfg_local_ip_list).");
		next;
	    }
	    if (exists $chk_ip{$cur_ip} || exists $active_session{$cur_ip}){
		#     IP    , .
		next;
	    }
	    $active_session{$cur_ip} = $cur_hostname;
	
	} elsif	($cur_line =~ /^.*$cfg_maillog_disconnect_mask$/){
	    sys_log(3, "DEBUG:DISCONNECT Line: $cur_line");

	    #  .
	    my $cur_ip = $2;
	    if ($cfg_log_format == 1){
		$active_session{$cur_ip} = $1;
	    }

	    if (check_ip_in_list(\@local_ip_bin_list, $cur_ip) == 1){
		#   IP.
		next;
	    }
	    #      " ".
	    if ((!exists $rejected_ip{$cur_ip}) && (exists $active_session{$cur_ip})){
	        #   .
	        my $cur_hostname = $active_session{$cur_ip};
	        if ($cfg_block_mask eq "" || $cur_hostname =~ /$cfg_block_mask/){
		    if ($cfg_ignore_mask eq "" || $cur_hostname !~ /$cfg_ignore_mask/){
		        #      " ".
		        $chk_ip{$cur_ip} = $active_session{$cur_ip};
		        $host_cnt++;
			
			if ($cfg_log_watch_mode == 1){
			    #     "tail" .
			    if ($now_time + $cfg_blocklist_rebuild_interval < time()){
				#   IP   .
				$now_time = time();
				if(check_spamer_in_ip_block(\%chk_ip, \%chk_cache) == 1){
				    #    ,     .
				    system("$cfg_blocklist_rebuild_cmd");
				    $db_obj->db_sync();
				}
				%chk_ip = ();
			    }
			}

		    } else {
    			sys_log(2, "INFO: $cur_ip/$cur_hostname ignored by rules cfg_ignore_mask.");
		    }
	    	} else {
    		    sys_log(2, "INFO: $cur_ip/$cur_hostname blocked by rules cfg_block_mask.");
		}
		delete $active_session{$cur_ip};
	    }else{ 
		if (exists $rejected_ip{$cur_ip}){
		    delete $rejected_ip{$cur_ip};
		}
		if (exists $active_session{$cur_ip}){
		    delete $active_session{$cur_ip};
		}
	    }

	} elsif ($cur_line =~ /^.*$cfg_maillog_reject_mask.*$/){
	    #   -  IP
	    my $cur_ip=$1;
	    sys_log(3, "DEBUG:REJECT Line: $cur_line");

	    if (check_ip_in_list(\@local_ip_bin_list, $cur_ip) == 1){
		#   IP.
		next;
	    }
	    $rejected_ip{$cur_ip} = 1;
	    #     -   ,    .
	    if (exists $active_session{$cur_ip}){
		my $cur_hostname = $active_session{$cur_ip};
    		sys_log(2, "INFO: Host $cur_ip/$cur_hostname already blocked.");
		delete $active_session{$cur_ip};
	    }
	}else{
	    sys_log(3, "DEBUG:IGNORED Line: $cur_line");
	}
    }
    $new_log_position = tell(MAILLOG);
    close(MAILLOG);

    #     .
    if ($cfg_log_watch_mode == 0){ # "cron"    .
	$new_log_size = (stat($cfg_maillog_file))[7];
	if ($new_log_position == -1){
	    sys_log(0, "ERROR:tell() return error value: $!");
	    $new_log_position = $new_log_size;
	}
	save_cur_position($new_log_position, $new_log_size, $cfg_seek_file);
    }

    #       .
    open (A_SES, ">$cfg_active_mail_file.new") or sys_log(-1, "ERROR:Can't store active sessions ($cfg_active_mail_file)");
    flock(A_SES, 2);
    foreach my $cur_ip (keys %active_session){
	my $cur_hostname = $active_session{$cur_ip};
	print A_SES "$cur_ip\t$cur_hostname\t$now_time\n";
    }
    close(A_SES);
    rename("$cfg_active_mail_file.new", $cfg_active_mail_file);

    #     .
    undef %active_session;
    undef %rejected_ip;


    if ($host_cnt == 0){
	#     , .
        untie %chk_cache;
	unlink($cfg_lock_file);
	exit(0);
    }

    if (($cfg_log_watch_mode == 0) 
	 && (check_spamer_in_ip_block(\%chk_ip, \%chk_cache) == 1)){
        #    ,     .
	system("$cfg_blocklist_rebuild_cmd");
    }

    $db_obj->db_sync();
    untie %chk_cache;
    unlink($cfg_lock_file);
    exit(0);
    
# END of main ------------------------------


##########################################################################
#        .
sub load_last_position{
    my ($position, $size, $seek_file) = @_;
    my ($str_data);
    $$position = 0;
    $$size = 0;
    if (-f "$seek_file"){
	open(LIST,"<$seek_file") or return -1;
        flock(LIST, 1);
	chomp($str_data = <LIST>);
	close(LIST);
	($$position, $$size) = split(/\t/, $str_data);
    }
    return 0;
}

##########################################################################
#           .
sub save_cur_position{
    my ($position, $size, $seek_file) = @_;

    open(LIST,">$seek_file") or return -1;
    flock(LIST, 2);
    print LIST "$position\t$size\n";
    close(LIST);
    return 0;
}
##########################################################################
#-------------------------------------------------------------------------
#     .
#-------------------------------------------------------------------------
sub sys_log{
	my($loglevel, $message) = @_;

    if ($loglevel > $cfg_verbose_level){
	return 0;
    }
     
    $message =~ s/[\n\r\|]/ /g;
    my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time());
    my $str_time = sprintf("%2.2d %2.2d %d %2.2d:%2.2d:%2.2d",$mday, $mon+1, 1900+$year, $hour, $min, $sec);
    if ($cfg_log_file eq "-"){
	#    STDOUT
        print "$str_time:$message\n";
    } else {
        open (SYSLOG, ">>$cfg_log_file") or return -1;
	flock (SYSLOG, 2);
        print SYSLOG "$str_time|$message\n";
	close (SYSLOG);
    }
    if ($loglevel == -1){
	unlink($cfg_lock_file);
	die $message;
    }
    return 0;
}

##########################################################################
#    .
sub add_to_blocklist{
    my ($cur_ip, $cur_hostname, $check_info) = @_;

    my $block_info="Blocked by spam_auto_checker. Host = $cur_hostname, Reason = $check_info";
    if (! open (BLOCKLIST, ">>$cfg_blocklist_file")){
	sys_log(0, "ERROR:Can't add rule to blacklist ($cfg_blocklist_file)");
	return -1;
    }
    flock(BLOCKLIST, 2);
    if ($cfg_blocklist_format == 1){
	# DSBL   ..
	print BLOCKLIST "$cur_ip:127.0.0.2:$block_info\n";
    } else {
        # "Access file" 
        print BLOCKLIST "$cur_ip\t $sendmail_format 550 $block_info\n";
    }
    close(BLOCKLIST);
    return 0;
}
##########################################################################
#     nmap,    
#    ,      root.
#       ,      
#       
#   .

sub check_remote_ports_nmap{
    my ($check_ip, $port_list, $open_ports) = @_;
    my $timeout_flag = 1;
    
    #   nmap.
    my $cfg_nmap_cmd="/usr/local/bin/nmap";
    
    if (! -x $cfg_nmap_cmd){
	sys_log(-1, "ERROR:nmap not found ($cfg_nmap_cmd)");
    }

    my $port_str = join(",", @$port_list);
    
    eval {
	    local $SIG{ALRM} = sub { die "timeout"; };
	    alarm(TIMEOUT);
	    open(NMAP, "$cfg_nmap_cmd -P0 --host_timeout 10000 --max_rtt_timeout 1000 --initial_rtt_timeout 1000 -p $port_str $check_ip|") || return -2;
	    while (<NMAP>){
		chomp();
		if (/^(\d+)\/\w+\s+open\s+.*$/){
		    push @$open_ports, $1;
		}
		if (/due to host timeout/){
		    $timeout_flag = -1;
		}
	    }
	    close(NMAP);
	    if ($timeout_flag == 1){
		$timeout_flag = 0;
	    }
    	    alarm(0);
    };

    if ($timeout_flag != 0){
        # ,    .
        return -1;
    }
    return 0;
}

##########################################################################
#         .
#       ,      
#       

sub check_remote_ports{
    my ($check_ip, $port_list, $open_ports) = @_;
    my $timeout_flag = 1;
    my $timeout_cnt = 0;
    
    my $proto = getprotobyname("tcp");
    my $thataddr = inet_aton($check_ip) || return -2;
    my $sockaddr = 'S n a4 x8';
    my $con_timeout = int (&TIMEOUT/(1 + scalar @$port_list)) + 1;
    
    foreach my $cur_port (@$port_list){
	my $timeout_flag = 1;
        my $that = pack($sockaddr, &AF_INET, $cur_port, $thataddr);
        socket(SOCK, &AF_INET, &SOCK_STREAM, $proto) || return -2;
	eval {
	    local $SIG{ALRM} = sub { die "timeout"; };
	    alarm($con_timeout);
    	    if (connect(SOCK, $that)){
		push @$open_ports, $cur_port;
	    }
	    $timeout_flag = 0;
    	    alarm(0);
	};
        close(SOCK);
	if ($timeout_flag != 0){
	    $timeout_cnt++;
	}
    }
    if ($timeout_cnt == scalar @$port_list){
	#    .
	return -1;
    } else {
	return 0;
    }
}

##########################################################################
#      IP
sub calc_weigth{
    my ($check_ip, $check_hostname, $check_info) = @_;
    my $check_weight = 0;
    my @check_info_arr = ();


    # Regex      .
    foreach my $cur_mask (keys %{$cfg_spam_factor{"domain_mask"}}){
	if ($check_hostname =~ /$cur_mask/i){
	    $check_weight += $cfg_spam_factor{"domain_mask"}->{$cur_mask};
	    push @check_info_arr, " domain_mask = " . $cfg_spam_factor{"domain_mask"}->{$cur_mask};
	}
    }

    #       , 0 -   .
    my $cur_domain_level = $check_hostname;
    $cur_domain_level =~ s/[^.]//g;
    $cur_domain_level = length($cur_domain_level);
    if (exists $cfg_spam_factor{"domain_level"}->{$cur_domain_level}){
        $check_weight += $cfg_spam_factor{"domain_level"}->{$cur_domain_level};
	push @check_info_arr, " domain_level($cur_domain_level) = " . $cfg_spam_factor{"domain_level"}->{$cur_domain_level};
    } else {
	#    .
        $check_weight += $cfg_spam_factor{"domain_level"}->{"N"};
	push @check_info_arr, " domain_level($cur_domain_level,N) = " . $cfg_spam_factor{"domain_level"}->{"N"};
    }

    #       .

    my @port_list = keys %{$cfg_spam_factor{"open_ports"}};

    if ($#port_list == -1){
	sys_log	("WARN:open_ports not specifed in config.");
    } else {
	#     nmap.

	my @open_ports = ();
	if (check_remote_ports($check_ip, \@port_list, \@open_ports) < 0){
	    #     .
	    return -1;
	}
	
	#  .
	foreach my $cur_port (@open_ports){
    	    $check_weight += $cfg_spam_factor{"open_ports"}->{"$cur_port"};
	    push @check_info_arr, " port $cur_port = " . $cfg_spam_factor{"open_ports"}->{"$cur_port"};
	}
    }
    push @check_info_arr, " weight = " . sprintf ("%.2f", $check_weight);
    
    $$check_info = join(",", @check_info_arr);

    if ($check_weight < 0){
	#    .
	return 0;
    }else{
	return $check_weight;
    }
}

##########################################################################
#     IP .
# $chk_ip -       IP.
# $chk_cache -        IP.
#  : 0 -   , 1 -  .

sub check_spamer_in_ip_block{
    my ($chk_ip, $chk_cache) = @_;
    my $flag_spamblock = 0;

    #   IP          
    #      .     -  
    #   .

    foreach my $cur_ip (keys %$chk_ip){
	my $cur_hostname = $$chk_ip{$cur_ip};
	my $cur_num_ip = inet_aton($cur_ip);
	if (! exists $$chk_cache{$cur_num_ip}){
	    # IP   ,     .
	    my $cur_weight = 0;
	    my $check_info = "";

	    #      IP
	    $cur_weight = calc_weigth($cur_ip, $cur_hostname, \$check_info);

	    if ($cur_weight == -1){
		sys_log(0, "TIMEOUT:$cur_ip:$cur_hostname:Host not alive");

	    }elsif ($cur_weight == -2){
		sys_log(0, "ERROR:$cur_ip:$cur_hostname:Null respond from checker.");

	    } elsif ($cur_weight < $cfg_spam_factor{"block_weight"}){
		#  .
		$$chk_cache{$cur_num_ip} = sprintf("%d:1:%.2f", $now_time, $cur_weight);
		sys_log(2, "OK:$cur_ip:$cur_hostname:$check_info");

	    } elsif ($cur_weight >= $cfg_spam_factor{"block_weight"}){
		# ,    .
		$$chk_cache{$cur_num_ip} = sprintf("%d:1:%.2f", $now_time, $cur_weight);
		$flag_spamblock = 1;		
		add_to_blocklist($cur_ip, $cur_hostname, $check_info);
		sys_log(1, "SPAM:$cur_ip:$cur_hostname:$check_info");
	    }
	} else {
    	    sys_log(2, "INFO: Host $cur_ip/$cur_hostname was checked before.");
	    #   .
	    if ($cfg_spam_factor{"reconnect_mul"} > 0){
		#  .
    	        my ($cur_time, $cur_con, $cur_weight) = split(/\:/, $$chk_cache{$cur_num_ip});
		$cur_con++;
		if (($cfg_spam_factor{"reconnect_mul"} != 1) 
		    && ($cur_weight < $cfg_spam_factor{"block_weight"})
		    && ($cur_time + 24*60*60 > $now_time)){
		    #       .
		    $cur_weight = $cur_weight * $cfg_spam_factor{"reconnect_mul"};
	    	    if ($cur_weight >= $cfg_spam_factor{"block_weight"}){
			# ,    .
			$flag_spamblock = 1;
			add_to_blocklist($cur_ip, $cur_hostname, "Connections per day = $cur_con, Weight = $cur_weight");
		    }
		}
		$$chk_cache{$cur_num_ip} = sprintf("%d:%d:%.2f", $cur_time, $cur_con, $cur_weight);
	    }
	}
    }
    return $flag_spamblock;
}

###############################################################################
#    x.x.x.x/y   .

use constant ADDR_NUM => 0;
use constant MASK_NUM => 1; 

sub optimize_ip{
    my ($ip_str) = @_;
    my @ip_bin=();
    my ($network_ip, $network_mask) = split(/\//, $ip_str);
    if ($network_mask == 0){ 
	$network_mask = 32 
    }
    $network_mask = pack ('B32', (1 x $network_mask), (1 x (32 - $network_mask)));
    $ip_bin[&ADDR_NUM] = inet_aton($network_ip);
    $ip_bin[&MASK_NUM] = $network_mask;
    return @ip_bin
}

#####################################################
#       .
#####################################################
sub optimize_ip_array{
    my ($ip_str, $ip_bin) = @_;
    my $i=0;

    foreach (@$ip_str){
	my ($network_ip, $network_mask) = optimize_ip($_);
	$$ip_bin[$i][&ADDR_NUM] = $network_ip;
	$$ip_bin[$i][&MASK_NUM] = $network_mask;
	$i++;
    }
    return 0;
}

#####################################################
#   IP  .
#####################################################
sub check_ip_in_subnet{
	my ($subnethost, $subnetmask, $ip) = @_;

    return (($subnethost & $subnetmask) eq ($ip & $subnetmask));
}

#####################################################
#   IP    .
#####################################################
# 1 - IP   , 0 - .
sub check_ip_in_list{
	my ($subnet_list, $ip) = @_;

    my $ip_bin = inet_aton($ip);
    foreach my $cur_local_net (@$subnet_list){
	if ( check_ip_in_subnet($$cur_local_net[&ADDR_NUM], $$cur_local_net[&MASK_NUM], $ip_bin) == 1){
	    return 1;
	}
    }
    return 0;
}
